context = None

import pygame

import random

mp_recovery_period = 5 * 10

try:
    import server
    context = "server"
    import monster
except ImportError:
    try:
        import client
        context = "client"
    except ImportError:
        raise

class Tile:

    def __init__(self,id):
        self.id = id

    def __repr__(self):
        return 'Tile(%s)' % repr(self.id)

    # a more compact representation for making a text representation of the map
    def __str__(self):
        if self.id < 50:
            return ' '
        else:
            return '#'

class Thing:
    def __init__(self, id, position, type = 'generic'):
        self.id = id
        self.position = position
        self.type = type

        self.debug = 0

    def __repr__(self):
        return ('Thing(%s, %s, %s)' %
                (repr(self.id), repr(self.position), repr(self.type)))

class Creature(Thing):

    def __init__(self, id, position, type = 'generic', hp = 20, max_hp = 21, mp = 15, max_mp = 15, bag = None):
        Thing.__init__(self, id, position, type)
        self.hp = hp
        self.mp = mp
        self.max_hp = max_hp
        self.max_mp = max_mp
        self.speed = 5
        self.visibility = 5
        if not bag:
            self.bag = {}
        else:
            self.bag = bag

        self.strength = 5
        self.dex = 5
        self.spirit = 5
        self.armor = 5

        # internal representation of self.engaged()
        self._engaged = {}
        if context == "server":
            self.check_engagement()

    # list of Creatures self is engaged with
    def engaged(self):
        return self._engaged.keys()
        
    def attack(self,creature):
        result = random.randint(1,10)
        if result == 10: return 0
        dex_mod = 0
        str_mod = 0
        arm_mod = 0
        if isinstance(self,Player):
            for each in self.bag.values():
                dex_mod += each.dex
                str_mod += each.strength
        if isinstance(creature,Player):
            for each in creature.bag.values():
                arm_mod += each.armor
        if result == 1 or result < self.dex + dex_mod:
            damage = max(int(float(random.randint(75,100))/100 * (self.strength + str_mod - creature.armor - arm_mod)),1)
            creature.hit(damage)
            return damage
        return 0

    def heal(self, level, creature):
        if self.mp < 4: return
        self.mp -= 4
        spr_mod = 0
        if isinstance(self, Player):
            for each in self.bag.values():
                spr_mod += each.spirit
        recovery = max(int(float(random.randint(75, 100))/100 * (self.spirit + spr_mod)), 1) * level
        creature.hit(-recovery)

    def spell(self, level, creature):
        if self.mp < 3: return 0
        self.mp -= 3
        attacker_spr_mod = 0
        defender_spr_mod = 0
        if isinstance(self, Player):
            for each in self.bag.values():
                attacker_spr_mod += each.spirit
        if isinstance(creature, Player):
            for each in creature.bag.values():
                defender_spr_mod += each.spirit
        damage = max(int(float(random.randint(75, 100)) / 100 * (self.spirit + attacker_spr_mod - creature.spirit - defender_spr_mod)), 1) * level
        creature.hit(damage)
        return damage

    def isDead(self):
        if self.hp <= 0:
            return True
        else:
            return False

    def hit(self, damage):
        self.hp = min(self.hp - damage, self.max_hp)
        if self.isDead():
            self.die()

    def die(self):
        if self.debug:
            print "%s) died" % self.id
        self.check_engagement()

    def check_engagement(self):
        # clear old engagement list
        for thing in self._engaged.keys():
            del thing._engaged[self]
        self._engaged = {}

        if self.isDead(): return
        
        x,y = self.position
        for i in (x-1,x,x+1):
            for j in (y-1,y,y+1):
                # XXX also checks self location,
                # but since no one else on same tile, don't worry about it
                if context == "client":
                    thing = client.positions.get((i,j))
                else:
                    thing = server.map.positions.get((i,j))
                if thing and self.is_enemy(thing):
                    self._engaged[thing] = 1
                    thing._engaged[self] = 1

    def items_in_bag(self):
        return self.bag

    def pick_item(self, new_item):
        position = new_item.position
        # swap out item of same type if exists in bag
        # old_item holds the older item which we will place into the map again
        if self.bag.has_key(new_item.type):
            old_item = self.bag[new_item.type]
        else:
            old_item = None
        self.bag[new_item.type] = new_item
        # remove new_item off the map
        del server.map.things[new_item.id]
        del server.map.positions[new_item.position]
        new_item.position = None
                
        # drop old_item to map
        # place old_item at new_item location
        if old_item is not None:
            old_item.position = position
            server.map.things[old_item.id] = old_item
            server.map.positions[old_item.position] = old_item

    def is_enemy(self, other):
        return False #non-specific creatures have no enemies

    def __repr__(self):
        return ('Creature(%s, %s, %s, %s, %s, %s, %s, %s)' %
                (repr(self.id), repr(self.position), repr(self.type), repr(self.hp), repr(self.max_hp), repr(self.mp), repr(self.max_mp), repr(self.bag)))

class Player(Creature):

    def __init__(self, id, position, type = 'generic', hp = 25, max_hp = 25, mp = 15, max_mp = 15, bag = None, crystals = None):
        Creature.__init__(self, id, position, type, hp, max_hp, mp, max_mp, bag)
        self.hp = hp
        self.max_hp = max_hp

        if not crystals:
            self.crystals = [] 
        else:
            self.crystals = crystals

        if not bag:
            self.bag = {}
        else:
            self.bag = bag

        self.mp_recovery_countdown = mp_recovery_period

    def own( self, crystal ):
        unique = {}
        for x in self.crystals:
            unique[x] = "hi"
        if ( not unique.has_key( crystal.id ) ):
            self.crystals.append( crystal.id )

    def get_crystal_count(self):
        return len( self.crystals )

    def die(self):
        Creature.die(self)
        if context == "server":
            for each in server.map.players:
                if not each.isDead(): return
            server.map.who_won = 'dm'

    def is_enemy(self, other):
        return isinstance(other, Monster) and not other.isDead()
        
    def flail(self): #XXX boring attack function
        if self.engaged():
            self.engaged()[0].hit(3)

    def maybe_recover_mp(self):
        if self.mp >= self.max_mp: return
        self.mp_recovery_countdown -= 1
        if self.mp_recovery_countdown <= 0:
            self.mp = min(self.mp + 5, self.max_mp)
            self.mp_recovery_countdown = mp_recovery_period

    def __repr__(self):
        crystal_list = []
        for x in self.crystals:
           crystal_list.append( x ) 

        return ('Player(%s, %s, %s, %s, %s, %s, %s, %s, %s)' %
                (repr(self.id), repr(self.position), repr(self.type), repr(self.hp), repr(self.max_hp), repr(self.mp), repr(self.max_mp), repr(self.bag), repr( crystal_list ) ))

class Monster(Creature):

    def __init__(self, id, position, type = 'generic', hp = 10, max_hp = 21, mp = 15, max_mp = 15):
        Creature.__init__(self, id, position, type, hp, max_hp, mp, max_mp)
        self.aggression = 64
        # stay put to start with
        self.wander_rectangle = pygame.Rect(position, (0, 0))
        self.paths = None
        self.pursuing = None
        # a pair (monster, position) of a monster that bumped this position
        # and that this monster should yield to
        self.rude_monster = None
        self.rank = 1 # precedence for yielding to other monsters

    def maybe_move(self):
        if self.isDead(): return False

        new_position = self.politeness(True)
        if new_position:
            if self.debug:
                print "%s) yield %s" % (self.id, repr(new_position))
            return new_position

        new_position = self.pursue()
        if new_position:
            if self.debug:
                print "%s) pursue %s" % (self.id, repr(new_position))
            return new_position
        
        new_position = self.follow_path()
        if new_position:
            if self.debug:
                print "%s) path %s" % (self.id, repr(new_position))
            return new_position

        # maybe be polite anyway if it has nothing better to do
        new_position = self.politeness(False)
        if new_position:
            if self.debug:
                print "%s) be polite %s" % (self.id, repr(new_position))
            return new_position

        new_position = self.wander()
        if new_position:
            return new_position
        
        return False

    # @param stubborn: True IFF the monster might have something better to do
    #                  something better to do
    def politeness(self, stubborn = False):
        if not self.rude_monster: return False
        rude_monster, position = self.rude_monster
        if stubborn and ((self.rank, self.id) >
                         (rude_monster.rank, rude_monster.id)):
            return False
        # if the monster already moved somewhere, no need to yield
        if rude_monster.position != position:
            self.rude_monster = None
            return False
        choices = []
        for delta_x in (-1, 0, 1):
            for delta_y in (-1, 0, 1):
                x = self.position[0] + delta_x
                y = self.position[1] + delta_y
                if ((delta_x, delta_y) != (0, 0) and
                    (x, y) != position and
                    server.map.hopefully_movable_space(x, y)):
                    choices.append((x, y))
        if not choices: return False
        choice = random.choice(choices)
        return choice[0] - self.position[0], choice[1] - self.position[1]

    def pursue(self):
        view_rect = pygame.Rect((self.position[0] - self.visibility,
                                 self.position[1] - self.visibility),
                                (2 * self.visibility, 2 * self.visibility))
        view_rect = server.map.rect.clip(view_rect)

        # prefer to continue pursuing the same enemy
        if self.pursuing and view_rect.collidepoint(self.pursuing.position):
            new_direction = self.pursue_target(self.pursuing)
            if new_direction:
                if self.debug:
                    print "%s) pursuing %s" % (self.id, self.pursuing)
                return new_direction
            
        for x in range(view_rect.left, view_rect.right + 1):
            for y in range(view_rect.top, view_rect.bottom + 1):
                try: thing = server.map.positions[(x, y)]
                except KeyError: continue
                if thing is self.pursuing: continue # already checked that
                if not isinstance(thing, Creature): continue
                new_direction = self.pursue_target(thing)
                if new_direction:
                    if self.debug:
                        print "%s) pursuing %s" % (self.id, thing)
                    self.pursuing = thing
                    return new_direction

        self.pursuing = None
        return False

    def pursue_target(self, target):
        if not self.is_enemy(target): return False
        paths = monster.ShortestPaths(pygame.Rect(target.position, (0, 0)))
        # XXX don't let the path go outside the pursuit rectangle
        distance, next = paths.search(self.position, monster.max_pursuit_path)
        if next == (None, None): return False
        if not self.pursuit_rectangle().collidepoint(next[0], next[1]):
            return False

        return (next[0] - self.position[0], next[1] - self.position[1])

    def follow_path(self):
        if not self.paths: return False

        distance, next = self.paths.search(self.position)
        if next == (None, None): #arrived at destination
            return False

        return (next[0] - self.position[0], next[1] - self.position[1])

    def wander(self):
        if random.randint(1,10) == 10:
            choices = []
            for x in range(-1,2):
                for y in range(-1,2):
                    if ((x or y) and 
                        server.map.movable_space(self.position[0]+x,
                                                 self.position[1]+y) and 
                        self.wander_rectangle.collidepoint(self.position[0]+x,
                                                          self.position[1]+y)):
                        choices.append((x,y))
            if choices: return random.choice(choices)
            
        return False

    def maybe_fight(self):
        if self.isDead(): return
        if random.randint(1,10) == 10:
            self.flail()

    def flail(self): #XXX boring attack function
        if self.engaged():
            self.engaged()[0].hit(1)

    def set_wander_rectangle(self, rectangle):
        if self.debug:
            print "%s) set wander rectangle to %s" % (self.id, repr(rectangle))
        self.wander_rectangle = server.map.rect.clip(rectangle)
        self.paths = monster.ShortestPaths(self.wander_rectangle)

    def pursuit_rectangle(self):
        return server.map.rect.clip(self.wander_rectangle.inflate(
              (self.aggression * 2, self.aggression * 2)))

    def is_enemy(self, other):
        return isinstance(other, Player) and not other.isDead()

    def die(self):
        Creature.die(self)
        if context == "server":
            del server.map.things[self.id]
            del server.map.positions[self.position]

    def __repr__(self):
        return ('Monster(%s, %s, %s, %s, %s, %s)' %
                (repr(self.id), repr(self.position), repr(self.type), repr(self.hp), repr(self.mp), repr(self.max_mp)))

class Item(Thing):

    contains = "chest"
    strength = 0
    dex = 0
    spirit = 0
    armor = 0

    def __init__(self, id, position, type = 'generic', contains = "chest", strength = 0, dex = 0, spirit = 0, armor = 0):
        Thing.__init__(self, id, position, type)
        self.contains = contains
        self.strength = strength
        self.dex = dex
        self.spirit = spirit
        self.armor = armor
    
    def __repr__(self):
        return ('%s(%s, %s, %s, %s, %s, %s, %s, %s )' %
                (self.__class__.__name__, repr(self.id), repr(self.position), repr(self.type), repr(self.contains), repr(self.strength), repr(self.dex), repr(self.spirit), repr(self.armor)))
                
class Armor(Item):
    pass
                
class Weapon(Item):
    pass
    
class Crystal(Item):
    def __init__(self, id, position, type = 'crystals', contains = "chest", color = 'blue', owner = None ):
        Item.__init__(self, id, position, type)
        self.color = color

        if ( color == "green" ):
            self.bmp = "green"
        elif ( color == "red" ):
            self.bmp = "red"
        elif ( color == "blue" ):
            self.bmp = "blue"
        elif ( color == "purple" ):
            self.bmp = "purple"

        self.owner = owner 
        self.owner_team = ''

    def get_image(self):
        return self.bmp
    
    def __repr__(self):

        #       name id  pos type  contain    color, owner
        return ('%s(%s, %s,  %s,   %s,        %s,     %s     )' %
                (self.__class__.__name__, repr(self.id), repr(self.position), repr(self.type), repr(self.contains), repr(self.color), repr(self.owner)) )

    def owned(self):
        return not self.owner == None
    
    def is_owned_by(self, person ):
        return self.owner == person.id

    def owned_by(self, person ):
        self.owner = person.id

    def get_color_hex(self):
        if ( self.color == 'green' ):
            return (0, 255, 0 )
        elif ( self.color == 'red' ):
            return ( 255, 0, 0 )
        elif ( self.color == 'blue' ):
            return ( 0, 0, 255 )
        elif ( self.color == 'purple' ):
            return ( 255, 0, 255 )


class Special(Item):
    pass


class Ogre(Monster):

    def __init__(self, id, position, type = 'ogre', hp = 30, max_hp = 30):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 7
        self.dex = 3
        self.spirit = 0
        self.armor = 3      


class Dragon(Monster):

    def __init__(self, id, position, type = 'dragon', hp = 50, max_hp = 50):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 9
        self.dex = 5
        self.spirit = 5
        self.armor = 5    


class Demon(Monster):

    def __init__(self, id, position, type = 'demon', hp = 35, max_hp = 35):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 5
        self.dex = 6
        self.spirit = 6
        self.armor = 3      


class CrystalMaster(Monster):

    def __init__(self, id, position, type = 'crystal', hp = 20, max_hp = 20):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 3
        self.dex = 3
        self.spirit = 7
        self.armor = 4


class Goblin(Monster):

    def __init__(self, id, position, type = 'goblin', hp = 15, max_hp = 15):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 4
        self.dex = 4
        self.spirit = 0
        self.armor = 3      


class Wolf(Monster):

    def __init__(self, id, position, type = 'wolf', hp = 10, max_hp = 10):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 3
        self.dex = 6
        self.spirit = 0
        self.armor = 2      


class Python(Monster):

    def __init__(self, id, position, type = 'python', hp = 5, max_hp = 5):
        Monster.__init__(self, id, position, type, hp, max_hp)
        self.strength = 2
        self.dex = 4
        self.spirit = 0
        self.armor = 0      


